<?php /** @noinspection PhpUndefinedMethodInspection */

/**
 * Carerix Wordpress Plugin
 *
 * @category settings
 * @author Bogdan Ivanov
 * @copyright Copyright (c) 2013 Carerix.com (http://www.carerix.com)
 * @link http://www.carerix.com
 * @since Aug 29, 2013 1:08:16 PM
 * @version $Rev$
 */

use Carerix\Api\Rest\Client;
use Carerix\Api\Rest\Entity;
use Carerix\Api\Rest\Entity\CRDataNode;
use Carerix\Api\Rest\Entity\CRPublication;
use Carerix\Api\Rest\Manager;


/**
 * Job class
 * Used to ....
 */
class Entity_Cron_Job extends Entity_Cron {

	protected $show = array('publicationID','publicationStart','creationDate', 
        'applicationContactInformation', 
		'companyInformation', 'functionContactInformation', 'introInformation',
		'layoutInstruction', 'notes', 'offerInformation', 'publicationSpecifications',
		'requirementsInformation', 'titleInformation', 'vacancyInformation', 'workLocation',
		'toMedium.code','toVacancy.vacancyID', 'toVacancy.jobTitle', 'publicationStart',
		);
	private $post_id = 0;
	protected $registerLanguage = false;
	private $modified_datetime = array();
	private $retention_buffer_days = 3;

	public function getUrl() {
		$jobsUrl = Environment_Installation::get_publications_feed();

		$queryString = $this->sourceData['settings']['feed'] ?? "";
		parse_str($queryString, $urlParamsArr);
		$urlParamsArr['medium'] = $urlParamsArr['medium'] ?: 'web';
		$jobsUrl .= '?' . http_build_query($urlParamsArr);

//		if( isset($this->sourceData['settings']['feed']) )
//		{
//			parse_str($this->sourceData['settings']['feed'], $urlParamsArr);
//			$urlParamsArr['medium'] = $urlParamsArr['medium'] ?: 'web';
//			$jobsUrl .= '?' . http_build_query($urlParamsArr);
//		}
//
//		if(stripos($jobsUrl, 'medium') === FALSE)
//		{
//			$extra = array('medium' => 'web');
//			$jobsUrl = $this->interfacer->url_merger($jobsUrl, $extra);
//		}

		if(    $this->registerLanguage 
			&& $this->interfacer->wpml_is_active()
            && strtolower(trim($this->registerLanguage)) != trim($this->interfacer->get_default_site_language()))
		{
			$jobsUrl = $this->interfacer->url_medium_for_translation($jobsUrl, $this->registerLanguage);
		}

		return $jobsUrl;
	}

	/**
	 * @param int $start
	 * @param int $count
	 *
	 * @return array
	 */
	private function getUrlQueryParts($start, $count)
	{
		$url   = $this->getUrl();
		$glue  = '?';
		$token = Environment_Settings::getAppToken();
		if(stripos($url, '?') !== FALSE) $glue = '&';
		$newUrl = $url . $glue . 'start=' . $start . '&count=' . $count . '&token='.$token;

		$parts = parse_url($newUrl);
		parse_str($parts['query'], $query);

		return $query;
	}

	public function getMediumCode()
	{
		$query = $this->getUrlQueryParts(0,0);
		$cx_mediumcode = $query['medium'];
		return $cx_mediumcode;
	}
	
	public function setRegisterLanguage($language)
	{
		$this->registerLanguage = $language;
	}


	public function retrieveData()
	{
		if ( Environment_Settings::getRestApiSync() == "on" ) {
			$feedData = $this->retrieveDataByRestApi(); // retrieve job items by the REST API
		} else {
			$feedData = parent::retrieveData(); // retrieve items (jobs, candidates, people, etc.) from a wp_feed
		}

		return $feedData;
	}


	/**
	 * Retrieve an array of inactive/deleted publication-IDs since a given date/time
	 *
	 * @param $reference_datetime
	 * @param $retention_buffer_days
	 *
	 * @return array
	 */
	public function retrieveInactivePublicationIdsByRestApi( $reference_datetime, $retention_buffer_days )
	{
		if(!$this->usesUrl) return null;

		$older_reference_datetime=date_sub(date_create($reference_datetime),date_interval_create_from_date_string("{$retention_buffer_days} days"));
		$older_reference_datetime = $older_reference_datetime->format("Y-m-d H:i:s");

//		$url = $this->getUrl();
//		$glue = '?';
//		$token = Environment_Settings::getAppToken();
//		if(stripos($url, '?') !== FALSE) $glue = '&';
//		$newUrl = $url . $glue . 'start=0&count=9999&token='.$token;
//
//		$parts = parse_url($newUrl);
//		parse_str($parts['query'], $query);

		$query = $this->getUrlQueryParts(0, 9999);

		$cx_start = $query['start'];
		$cx_count = $query['count'];
		$cx_mediumcode = $query['medium'];

		$app_name = Environment_Installation::getAppName();
		$app_token = Environment_Installation::getAppToken();

		$timezone_as_emailformat = $this->formatTimeZoneAsEmail(); // something like +0100 or +0200
		$reference_date_as_nscalendardate = "(NSCalendarDate)'{$reference_datetime} {$timezone_as_emailformat}'";
		$older_reference_date_as_nscalendardate = "(NSCalendarDate)'{$older_reference_datetime} {$timezone_as_emailformat}'";
		$end_of_today_as_nscalendardate = "(NSCalendarDate)'" . date("Y-m-d") . " 23:59:59 {$timezone_as_emailformat}'";
		$start_of_today_as_nscalendardate = "(NSCalendarDate)'" . date("Y-m-d") . " 00:00:00 {$timezone_as_emailformat}'";

		$arrPubIDs = array();

		// http://development.wiki.carerix.com/cxwiki/doku.php?id=cxrest_api_client
		try
		{
			$client  = new Client();
			$manager = new Manager($client);
			$manager->autoDiscoverEntities();
			$manager->setUrl(WP_Constants::CARERIX_API_ENDPOINT);
			$manager->setUsername($app_name);
			$manager->setPassword($app_token);

			Entity::setManager($manager);

			/**
			 * Fetch publications which qualify to be deleted from Wordpress
			 *
			 * General conditions are:
			 * 1. filter on the medium code
			 * 2. it's not an open applicaton
			 *
			 * Conditional logic:
			 * - publicaton or vacancy has status 'deleted' and the modification date is up to `retention_buffer_days` in the past
			 * or
			 * - publication end-date or materials deadline-date is up to `retention_buffer_days` in the past
			 **/
			$medium_clause = $this->interfacer->constructSqlMediumClause( $cx_mediumcode );
//			$medium_clause = $this->__constructSqlMediumClause( $cx_mediumcode );

			$qualifier = <<<HEREDOC
(
  {$medium_clause}
) 
and 
(
  toVacancy.isOpenApplication != 1
  or
  toVacancy.isOpenApplication = nil
) 
and 
(
  (
	deleted = 1
	and
    (
      modificationDate > {$older_reference_date_as_nscalendardate}
      or
      toVacancy.modificationDate > {$older_reference_date_as_nscalendardate}
    )
  ) 
  or
  (
    (
      publicationEnd != nil
      and
      publicationEnd > {$older_reference_date_as_nscalendardate}
	  and
	  publicationEnd < {$start_of_today_as_nscalendardate}
	)
	or
	(
	  deadlineMaterial != nil
	  and
      deadlineMaterial > {$older_reference_date_as_nscalendardate}
	  and
	  deadlineMaterial < {$start_of_today_as_nscalendardate}
	)
	or
	(
	  publicationStart != nil
	  and
      publicationStart > {$end_of_today_as_nscalendardate}
	)
  )
)
HEREDOC;

//			$qualifier = "(";
//			$qualifier .= " toMedium.code = '{$cx_mediumcode}'";
//			$qualifier .= ") and (";
//			$qualifier .= " toVacancy.isOpenApplication != 1";
//			$qualifier .= " or toVacancy.isOpenApplication = nil";
//			$qualifier .= ") and (";
//			$qualifier .= " modificationDate >= {$reference_date}"; // entity publication
//			$qualifier .= " or toVacancy.modificationDate >= {$reference_date}"; // entity joborder
//			$qualifier .= ") and (";
//			$qualifier .= " deleted = 1"; // entity publication
//			$qualifier .= " or toVacancy.deleted = 1"; // entity joborder
//			$qualifier .= " or publicationStart > (NSCalendarDate)'".date("Y-m-d")." 23:59:59 {$timezone_as_emailformat}'";
//			$qualifier .= " or publicationEnd < (NSCalendarDate)'".date("Y-m-d")." 00:00:00 {$timezone_as_emailformat}'";
//			$qualifier .= ")";


			$params = array(
				'start' => $cx_start,
				'count' => $cx_count,
				'qualifier' => $qualifier,
				'norestrict' => 1,
				'show' => array('publicationID'),
			);
			$data = @CRPublication::findAll($params);

			$arrPublications = $data->getArrayCopy();

			foreach ($arrPublications as $publication)
			{
				$arrPubIDs[] = $publication->getPublicationID();
			}
		}  catch (Carerix\Api\Rest\Exception $care) {
			// print_r($care);
			die();
		}
		return $arrPubIDs;
	}

	/**
	 * @return SimpleXMLElement|null
	 * @throws Exception
	 */
	private function retrieveDataByRestApi()
	{
		$xml_bootstrap = '<rss version="0.91"><channel></channel></rss>';
		// $feedData = simplexml_load_string($xml_bootstrap);
		$feedData = new SimpleXMLElement($xml_bootstrap);

		if(!$this->usesUrl) return null;
		$page = $this->page;
		$start = (int)$page * WP_Constants::NR_OF_ITEMS;

		$query = $this->getUrlQueryParts($start, WP_Constants::NR_OF_ITEMS);

		$cx_start = $query['start'];
		$cx_count = $query['count'];
		$cx_mediumcode = $query['medium'];
		$cx_forced_lang_name = $query['language'] ?? false; // language=Xxxxx is manually added to the "job feed source" configuration

		$interfacer = WP_Interfacer::getInstance();
//		$cx_forced_language_code = $interfacer->convertLanguageNameToCode($cx_forced_lang_name);
//		print_r($cx_forced_language_code);
		$lang_obj = $interfacer->extractLanguageFromMediumCode( $cx_mediumcode );

		$app_name = Environment_Installation::getAppName();
		$app_token = Environment_Installation::getAppToken();
		$additionalfieldssupport_status = Environment_Settings::getCxAdditionalFieldsSupportStatus(); // `on` or `off`


		$client = new Client();
		$manager = new Manager($client);
		$manager->autoDiscoverEntities();
		$manager->setUrl(WP_Constants::CARERIX_API_ENDPOINT);
		$manager->setUsername($app_name);
		$manager->setPassword($app_token);
		Entity::setManager($manager);


		// https://api.carerix.com/CRDataNode/3661?show=value&show=exportCode&show=tag
		// https://api.carerix.com/CRDataNode/list-by?type=Attribuut%20publicatie&show=value&show=exportCode&show=tag&show=modificationDate&language=English

		$arrExtraFieldsDictionary = array();
		if ($additionalfieldssupport_status == "on") {
			// tagtypes: listType,checkboxListType,stringType,etc.
			$params = array(
				/*'norestrict' => 1,*/ // disrespect notActive?
				'type'     => 'Attribuut publicatie',
				'show'     => array('isHidden', 'value', 'exportCode', 'tag', 'modificationDate', 'type.name'),
				'language' => $cx_forced_lang_name ?: $lang_obj["name"],
				'_cxwp'    => Carerix::VERSION
			);

			$attributes = array("Attribuut publicatie", "Attribuut vacature");
			foreach ($attributes as $attribute) {
				$params['type'] = $attribute;
				$data    = CRDataNode::listBy($params);
				$arrData = $data->getArrayCopy();

				/* @var $crdatanode CRDataNode */
				foreach ($arrData as $crdatanode)
				{
					$exportCode                        = trim((string) $crdatanode->getExportCode());

					// exportCode is required. Though required, if the exportCode-qualifyer is removed, the system will alternatively (!!!!) use dataNodeID. In that case the 'WP vacancy post fieldname' will be something like 'cx_extra_1276' instead of 'cx_extra_myfieldname'
					if ( strlen($exportCode) == 0 ) continue;
					$nodeID                            = $crdatanode->getDataNodeID();
					$arrExtraFieldsDictionary[$nodeID] = $crdatanode;
					$this->__addModificationDate($crdatanode->modificationDate);
				}
			}

		} // endif


		// http://development.wiki.carerix.com/cxwiki/doku.php?id=cxrest_api_client
		try {
			$tables = "Land|Regio|Branche0|Branche1|Functie0|Functie1|Opleiding0|Opleiding1|Opleiding2|Groep-Vacature|Functieniveau|Dienstverband|Salarisperiode|Valuta|Productgroep";
			$arrTables = explode("|", $tables);
			$sqlClause = "";
			foreach ($arrTables as $table)
			{
				$sqlClause .= "type.name=\"{$table}\"";
				if ( $table !== end($arrTables) ) {
					$sqlClause .= " OR ";
				}
			}

			$params  = array(
				'qualifier' => "notActive != 1 and deleted != 1 and ($sqlClause)",
				'show'      => array('values.modificationDate', 'values.value'),
				'ordering'  => "({key=values.modificationDate;sel=Descending})",
				'count'     => 1,
				'_cxwp'     => Carerix::VERSION
			);

			$data = wp_cache_get( 'restapi_tables_mod_date', CARERIX_CACHE_GROUP );
			if ( false === $data ) {
				$data = @CRDataNode::findAll($params);
				wp_cache_set( 'restapi_tables_mod_date', $data, CARERIX_CACHE_GROUP, 60 );
			}
			$arrData = $data->getArrayCopy();


			// Get newest modification date of changes from the CX tables involved (var $tables)
			foreach ($arrData as $itemList) {
				$items = $itemList->values;
				foreach ($items as $item) {
					if (isset($item->modificationDate)) $this->__addModificationDate( $item->modificationDate);
				}
			}


			$source_id = @$this->sourceData['id'];
			$sync_reference_date = $this->getCxIsSyncedToThisDateTime($source_id);


			if (strtotime($this->getCxTablesLastModificationDate()) > strtotime($sync_reference_date)) {
				$sync_reference_date = null; // do a full sync!
			}

			$qualifier = $this->interfacer->getPublicationsListQualifier( $cx_mediumcode, $sync_reference_date);

//			$timezone_as_emailformat = $this->formatTimeZoneAsEmail(); // something like +0100 or +0200
//			$reference_date_as_nscalendardate = "(NSCalendarDate)'{$sync_reference_date} {$timezone_as_emailformat}'";
//			$older_reference_datetime=date_sub(date_create($sync_reference_date),date_interval_create_from_date_string("{$this->retention_buffer_days} days"));
//			$older_reference_datetime = $older_reference_datetime->format("Y-m-d H:i:s");
//			$older_reference_date_as_nscalendardate = "(NSCalendarDate)'{$older_reference_datetime} {$timezone_as_emailformat}'";
//			$end_of_today_as_nscalendardate = "(NSCalendarDate)'" . date("Y-m-d") . " 23:59:59 {$timezone_as_emailformat}'";
//			$start_of_today_as_nscalendardate = "(NSCalendarDate)'" . date("Y-m-d") . " 00:00:00 {$timezone_as_emailformat}'";
//
//			$medium_clause = $this->__constructSqlMediumClause( $cx_mediumcode );

/*
			$qualifier = <<<HEREDOC
(
  {$medium_clause}
) 
and
(
  toVacancy.deleted != 1
)
and 
(
  toVacancy.isOpenApplication != 1
  or
  toVacancy.isOpenApplication = nil
) 
and
(
  (
	(
	  modificationDate >= {$reference_date_as_nscalendardate}
	  or toVacancy.modificationDate >= {$reference_date_as_nscalendardate}
	  or toMedium.modificationDate >= {$reference_date_as_nscalendardate}
	  or owner.modificationDate >= {$reference_date_as_nscalendardate}
	  or toVacancy.agency.modificationDate >= {$reference_date_as_nscalendardate}
	  or toVacancy.toCompany.modificationDate >= {$reference_date_as_nscalendardate}  
	  or toVacancy.toUser.modificationDate >= {$reference_date_as_nscalendardate}
	)
	and (publicationStart <= {$end_of_today_as_nscalendardate})
	and (publicationEnd = nil or publicationEnd > {$start_of_today_as_nscalendardate})
	and (deadlineMaterial = nil or deadlineMaterial > {$start_of_today_as_nscalendardate})
  )
  or
  (
    publicationStart > {$older_reference_date_as_nscalendardate}
	and (publicationEnd = nil or publicationEnd > {$start_of_today_as_nscalendardate})
	and (deadlineMaterial = nil or deadlineMaterial > {$start_of_today_as_nscalendardate})
  )
)
HEREDOC;
*/
			$params = array(
				'start' => $cx_start,
				'count' => $cx_count,
				'qualifier' => $qualifier,
				'language' => $cx_forced_lang_name ?: $lang_obj["name"],
				'show' =>   array(
					'titleInformation',
					'additionalInfo',
					'toMedium',
					'toVacancy',
					'toVacancy.toContractTypeNode',
					'toVacancy.toSalaryCurrencyNode',
					'toVacancy.toSalaryPeriodNode',
					'toVacancy.toSalaryPeriodNode.parentNodes',
					'toVacancy.groups',
					'toVacancy.groups.modificationDate',
					'toVacancy.publications.id',
					'toVacancy.toWorkLevelNode.value',
					'toVacancy.toWorkLevelNode.modificationDate',
					'toVacancy.toBrancheLevel1.value',
					'toVacancy.toBrancheLevel1.modificationDate',
					'toVacancy.toBrancheLevel2.value',
					'toVacancy.toBrancheLevel2.modificationDate',
					'toVacancy.toFunctionLevel1',
					'toVacancy.toFunctionLevel1.values',
					'toVacancy.toFunctionLevel2',
					'toVacancy.toFunctionLevel2.values',
					'toVacancy.jobTitle',
					/* 'introInformationHTML', */
					'introInformation',
					/*'companyInformationHTML',*/
					'companyInformation',
					/*'vacancyInformationHTML',*/
					'vacancyInformation',
					/*'offerInformationHTML',*/
					'offerInformation',
					/*'requirementsInformationHTML',*/
					'requirementsInformation',
					'workLocation',
					/* 'owner.informalName', */
					'owner.lessFormalNameString',
					'owner.modificationDate',
					'toVacancy.agency.name',
					'toVacancy.agency.modificationDate',
					/*'functionContactInformationHTML',*/
					'functionContactInformation',
					/*'applicationContactInformationHTML',*/
					'applicationContactInformation',
					'toVacancy.toCountry1Node',
					'toVacancy.toCountry1Node.modificationDate',
					'toVacancy.toProvince1Node',
					'toVacancy.toProvince1Node.modificationDate',
					'toVacancy.toCountry2Node',
					'toVacancy.toCountry2Node.modificationDate',
					'toVacancy.toProvince2Node',
					'toVacancy.toProvince2Node.modificationDate',
					'toVacancy.toCountry3Node',
					'toVacancy.toCountry3Node.modificationDate',
					'toVacancy.toProvince3Node',
					'toVacancy.toProvince3Node.modificationDate',
					'toVacancy.isAnonymous',
					'toVacancy.workCity',
					'toVacancy.workPostalCode',
					'toVacancy.invoicePostalCode',
					'toVacancy.toCompany.name',
					'toVacancy.toCompany.modificationDate',
					'toVacancy.toCompany.visitPostalCode',
					'toVacancy.educations.toLevel1Education1.value',
					'toVacancy.educations.toLevel1Education1.modificationDate',
					'toVacancy.educations.toLevel1Education2.value',
					'toVacancy.educations.toLevel1Education2.modificationDate',
					'toVacancy.educations.toLevel1Education3.value',
					'toVacancy.educations.toLevel1Education3.modificationDate',
					'toVacancy.toProductNode.value', /* procedure */
					'toVacancy.toProductNode.modificationDate',
					'toVacancy.toProductTypeNode.value', /* business line (vakgebied) */
					'toVacancy.toProductTypeNode.modificationDate',
					'toVacancy.toUser.firstName',
					'toVacancy.toUser.lastNamePrefix',
					'toVacancy.toUser.lastName',
					'toVacancy.toUser.modificationDate',
					'creationDate',
					'modificationDate',
					'publicationStart',
					'publicationEnd',
					'deadlineMaterial', // last date to apply by the publication (CX->Publication->Tab general->Closing Date)
					'toVacancy.reactionEndDate', // last date to apply as set in CX->Job order->Tab publication text->Apply before
					'toVacancy.vacancyID',
					'toVacancy.vacancyNo',
					'toVacancy.isOpenApplication',
					'toVacancy.creationDate',
					'toVacancy.modificationDate'
				),
				'ordering' => '({key=publicationStart;sel=Descending})', // sort descending by publication date
			);

//			$params = array(
//				'start' => 0,
//				'count' => 999,
//				'qualifier' => $qualifier,
//				'ordering' => '({key=publicationStart;sel=Descending})', // sort descending by publication date
//			);


			$jobs_finished_processing = wp_cache_get( "restapi_jobs_finished_processing_{$cx_mediumcode}", CARERIX_CACHE_GROUP );
			if ( false !== $jobs_finished_processing ) {
				$manager = null;
				unset($manager);
				$client = null;
				unset($client);
				return $feedData;
			}

			$data = @CRPublication::findAll("list", $params);

			$arrPublications = $data->getArrayCopy();
			$xmlPublications=simplexml_load_string($data->getRaw()) or die("Error: Cannot create object");
			$grand_total_publications = (int)$xmlPublications->attributes()->count;


			// No need to process more publications because there are none left for this source (=medium)
			if ( $cx_start >= $grand_total_publications - ($grand_total_publications % $cx_count)) {
				wp_cache_set( "restapi_jobs_finished_processing_{$cx_mediumcode}", 1, CARERIX_CACHE_GROUP, 10 ); // cache for 10 secs
			}


			/* Current memory usage  */
			$mem_usage = memory_get_usage();
			/* Peak memory usage */
			$mem_peak = memory_get_peak_usage();


			if ($this->manual_called) {
				if ($page == 0) {
					$onetime= wp_cache_get( 'restapi_one_time', CARERIX_CACHE_GROUP );
					if ( false === $onetime ) {
						wp_cache_set( 'restapi_one_time', 1, CARERIX_CACHE_GROUP, 10 );
						echo "\n<p>";
						echo "\nLast synchronization was up to: " . $sync_reference_date . "<br/>";
						echo "\nLast modification datetime of CX tables: " .  $this->getCxTablesLastModificationDate() . "<br/>";
						echo "\nCurrent server UTC datetime: " . current_time( "mysql", 1) . "<br/>";
						echo "\n</p>";
						echo "\n";
					}
				}

				echo "\n<p>";
				echo "\nPROCESSING NEW/MODIFIED JOBS for mediumcode `{$cx_mediumcode}` (page " . ($page+1) . " of " . (floor($grand_total_publications / $cx_count)+1) . ")<br/>";
				// echo "\nSyncing mediumcode `{$cx_mediumcode}` for `{$grand_total_publications}` jobs in total";
				echo "\nOffset start->end: {$cx_start}->" . ($cx_start+$cx_count) . "<br/>";
				echo "\nJobs processed: " . count($arrPublications) . "<br/>";
				echo "\nScript is now using: " . round($mem_usage / 1024 / 1024) . "MB of memory. (Peak usage: " . round($mem_peak / 1024 / 1024) . "MB)<br/>";
				echo "\n</p>";
				echo "\n";
			}

			$i=0;
			/* @var $publicaton CRPublication */
			foreach ($arrPublications as $publication)
			{
				$this->modified_datetime = array(); // clear object for each publication
				$publicationXML = &$xmlPublications->CRPublication[$i++];

				$thisItem = $feedData->channel->addChild("item");
				$thisItem->addChild("guid", (string) $publication->getPublicationID());

				@$this->__addModificationDate( (string) $publication->getModificationDate() );   // publication modification date
				@$this->__addModificationDate( (string) $publication->getToVacancy()->getModificationDate() ); // vacancy modification date

				$pubIdList = $this->getOtherPubIdsBelongingToThisPublication( $publication );
				$thisItem->addChild("pubIdList", $pubIdList);

				$pubDate = $this->formatPublicationDate( (string) $publication->getPublicationStart() );
				$thisItem->addChild("pubDate", $pubDate);

				$title = $this->formatPublicationTitle( $publication );
				$thisItem->addChild("title", $title);

				$thisItem->addChild("postalCode", htmlspecialchars((string) $publication->getToVacancy()->workPostalCode));
				$thisItem->addChild("rawTitle", htmlspecialchars(trim((string) $publication->getTitleInformation())));
				$owner = $thisItem->addChild("ownerName", htmlspecialchars((string) $publication->getOwner()->lessFormalNameString));
				$owner->addAttribute("id", (string) $publication->getOwner()->userID); // print_r() won't show it, but it's definately there!
				$this->__addModificationDate( (string) $publication->getOwner()->modificationDate );
				$agency = $thisItem->addChild("agency", htmlspecialchars((string) $publication->getToVacancy()->agency->name));
				$this->__addModificationDate( (string) $publication->getToVacancy()->agency->modificationDate );
				$agency->addAttribute("id", (string) $publication->getToVacancy()->agency->agencyID); // print_r() won't show it, but it's definately there!
				$thisItem->addChild("contact",""); // Unused. Removed because of GDPR?
				$thisItem->addChild("company", $this->publicationIsAnonymous($publication) ? "" : htmlspecialchars((string) $publication->getToVacancy()->getToCompany()->name));
				$this->__addModificationDate( (string) $publication->getToVacancy()->getToCompany()->modificationDate );
				$thisItem->addChild("startDate", (string) $publication->getPublicationStart());
				$thisItem->addChild("endDate", (string) $publication->getPublicationEnd());
				$thisItem->addChild("deadlineMaterial", $publication->getDeadlineMaterial());
				$thisItem->addChild("reactionEndDate", $publication->getToVacancy()->getReactionEndDate());

				$titleWorkLocation = $this->getTitleWorkLocation( $publication );
				$thisItem->addChild("titleWorkLocation", htmlspecialchars($titleWorkLocation));

				$thisItem->addChild("workLocation", (string) $publication->getWorkLocation());

				@$thisItem->addChild("country1", htmlspecialchars((string) $publication->getToVacancy()->getToCountry1Node()->value));
				@$this->__addModificationDate( (string) $publication->getToVacancy()->getToCountry1Node()->modificationDate );
				@$thisItem->addChild("region1", htmlspecialchars((string) $publication->getToVacancy()->getToProvince1Node()->value)); // Suppress with "@", because the system bugs on this once
				@$this->__addModificationDate( (string) $publication->getToVacancy()->getToProvince1Node()->modificationDate );
				@$thisItem->addChild("country2", htmlspecialchars((string) $publication->getToVacancy()->getToCountry2Node()->value));
				@$this->__addModificationDate( (string) $publication->getToVacancy()->getToCountry2Node()->modificationDate );
				@$thisItem->addChild("region2", htmlspecialchars((string) $publication->getToVacancy()->getToProvince2Node()->value));
				@$this->__addModificationDate( (string) $publication->getToVacancy()->getToProvince2Node()->modificationDate );
				@$thisItem->addChild("country3", htmlspecialchars((string) $publication->getToVacancy()->getToCountry3Node()->value));
				@$this->__addModificationDate( (string) $publication->getToVacancy()->getToCountry3Node()->modificationDate );
				@$thisItem->addChild("region3", htmlspecialchars((string) $publication->getToVacancy()->getToProvince3Node()->value));
				@$this->__addModificationDate( (string) $publication->getToVacancy()->getToProvince3Node()->modificationDate );

				$thisItem->introInformation = (string) $publication->introInformation;
				$thisItem->addChild("companyInformation", (string) $publication->companyInformation);
				@$thisItem->addChild("branche0", htmlspecialchars((string) $publication->getToVacancy()->getToBrancheLevel1()->value));
				@$this->__addModificationDate( (string) $publication->getToVacancy()->getToBrancheLevel1()->modificationDate );
				@$thisItem->addChild("branche1", htmlspecialchars((string) $publication->getToVacancy()->getToBrancheLevel2()->value));
				@$this->__addModificationDate( (string) $publication->getToVacancy()->getToBrancheLevel2()->modificationDate );
				@$thisItem->addChild("category", htmlspecialchars((string) $publication->getToVacancy()->getToFunctionLevel1()->value));
				@$thisItem->addChild("function0", htmlspecialchars((string) $publication->getToVacancy()->getToFunctionLevel1()->value));
				@$this->__addModificationDate( (string) $publication->getToVacancy()->getToFunctionLevel1()->modificationDate );
				@$thisItem->addChild("function1", htmlspecialchars((string) $publication->getToVacancy()->getToFunctionLevel2()->value));
				@$this->__addModificationDate( (string) $publication->getToVacancy()->getToFunctionLevel2()->modificationDate );
				@$thisItem->addChild("toWorkLevelNode", htmlspecialchars((string) $publication->getToVacancy()->getToWorkLevelNode()->value));
				@$this->__addModificationDate( (string) $publication->getToVacancy()->getToWorkLevelNode()->modificationDate );
				$thisItem->addChild("vacancyInformation", (string) $publication->vacancyInformation);
				$thisItem->addChild("requirementsInformation", (string) $publication->requirementsInformation);
				$thisItem->addChild("offerInformation", (string) $publication->offerInformation);
				$thisItem->addChild("functionContactInformation", (string) $publication->functionContactInformation);
				$thisItem->addChild("applicationContactInformation", (string) $publication->applicationContactInformation);
				$thisItem->addChild("vacancyID", (string) $publication->getToVacancy()->vacancyID);
				$thisItem->addChild("vacancyNo", (string) $publication->getToVacancy()->vacancyNo);
				$thisItem->addChild("medium", (string) $publication->getToMedium()->code);
				@$this->__addModificationDate( (string) $publication->getToMedium()->modificationDate );
				@$thisItem->addChild("procedure", htmlspecialchars((string) $publication->getToVacancy()->getToProductNode()->value));
				@$this->__addModificationDate( (string) $publication->getToVacancy()->getToProductNode()->modificationDate );
				@$thisItem->addChild("businessLine", htmlspecialchars((string) $publication->getToVacancy()->getToProductTypeNode()->value));
				@$this->__addModificationDate( (string) $publication->getToVacancy()->getToProductTypeNode()->modificationDate );
				$thisItem->addChild("hoursPerWeek", htmlspecialchars((string) $publication->getToVacancy()->hoursPerWeek));
				@$thisItem->addChild("contractType", htmlspecialchars((string) $publication->getToVacancy()->getToContractTypeNode()->value));

				$isSalaryPublic = (int) $publication->getToVacancy()->isSalaryPublic;
				$thisItem->addChild("isSalaryPublic", $isSalaryPublic);
				$thisItem->addChild("minSalary", $isSalaryPublic ? (string) $publication->getToVacancy()->minSalary : null);
				$thisItem->addChild("maxSalary", $isSalaryPublic ? (string) $publication->getToVacancy()->maxSalary : null);
				@$thisItem->addChild("salaryPeriod", $isSalaryPublic ? htmlspecialchars((string) $publication->getToVacancy()->getToSalaryPeriodNode()->value) : null);
				@list($objSalaryPeriodClassification) = $publication->getToVacancy()->getToSalaryPeriodNode()->getParentNodes();
				@$thisItem->addChild("salaryPeriodClassification", $isSalaryPublic ? (string) $objSalaryPeriodClassification->value : null);
				$thisItem->addChild("salaryCurrency", $isSalaryPublic ? htmlspecialchars((string) $publication->getToVacancy()->getToSalaryCurrencyNode()->value) : null);


				// Educations
				$arrEducations = $this->getEducations( $publicationXML );
				$educations = $thisItem->addChild("educations"); // placeholder tag
				foreach ($arrEducations as $education)
				{
					$educations->addChild("education", $education["education"]); // child of educations
					$educations->addChild("required", $education["required"]); // child of educations
				}


				// Vacancy groups
				$groups = $thisItem->addChild("groups"); // placeholder tag
				$arrGroups = $this->getPublicationGroups( $publicationXML );
				foreach ($arrGroups as $itemGroup)
				{
					$group = $groups->addChild("group", $itemGroup["value"]);
					$group->addAttribute("id", $itemGroup["id"]);
					$group->addAttribute("exportCode", $itemGroup["exportCode"]);
				}


				// Additional information
				if ( $additionalfieldssupport_status == "on" ) {
					$arrExtraFieldData_publication = $publication->getAdditionalInfo()->getArrayCopy();
					$arrExtraFieldData_vacancy     = $publication->getToVacancy()->getAdditionalInfo()->getArrayCopy();

					$additionalInfo = $thisItem->addChild("additionalInfo"); // placeholder tag

					foreach ( (array($arrExtraFieldData_vacancy, $arrExtraFieldData_publication)) as $arrExtraFieldData )
					{
						foreach ($arrExtraFieldData as $CRDataNodeId => $itemExtraFieldData)
						{
							try
							{
								$fieldtype  = $arrExtraFieldsDictionary[$CRDataNodeId]->getTag();// getType()->getName();
								$exportCode = $arrExtraFieldsDictionary[$CRDataNodeId]->getExportCode();
								$isHidden   = (int) $arrExtraFieldsDictionary[$CRDataNodeId]->isHidden;
							}
							catch (\Throwable $t)
							{
								$fieldtype  = "";
								$exportCode = "";
								$isHidden   = "";
							}

							if ($isHidden)
							{
								$fieldtype = "hiddenType";
							}
							$fieldvalue = array();
							// fieldtypes: campagnelijst, tabel item lijst, datum+tijdstip, wachtwoord, url, tekst,
							// tekst multiline, geheel getal, getal, datum, aankruisvakje, lijst, aankruisvakjes lijst,
							// radiobuttons lijst, gebruikerslijst, gebruikers van mijn vestiging, lijst vestigingen

							switch ($fieldtype)
							{
								case "integerType":
									$fieldvalue[] = (int) $itemExtraFieldData;
									break;

								case "numberType":
									$fieldvalue[] = (float) $itemExtraFieldData;
									break;

								case "dateType": // e.g. 2017-10-12T08:00:00Z+0200
									$itemExtraFieldData = date('Y-m-d', strtotime((string) $itemExtraFieldData));
									$fieldvalue[]       = $itemExtraFieldData;
									break;

								case "dateTimeType": // e.g. 2017-10-12T11:15:00Z+0200
									$itemExtraFieldData = date('Y-m-d H:i:s', strtotime((string) $itemExtraFieldData));
									$fieldvalue[]       = $itemExtraFieldData;
									break;

								case "checkboxType": // e.g. 1
									$fieldvalue[] = (int) $itemExtraFieldData;
									break;

								case "stringType":
								case "textareaType":
								case "passwordType":
								case "urlType":
									$fieldvalue[] = $itemExtraFieldData;
									break;

								case "listType": // single option
								case "radiobuttonListType": // single option
									$fieldvalue[] = CRDataNode::find($itemExtraFieldData)->getValue();
									break;

								case "checkboxListType": // multiple options possible
									/* @var $ns_array Carerix\Api\Rest\DataType\NSArray */
									$ns_array           = $itemExtraFieldData;
									$itemExtraFieldData = $ns_array->getArrayCopy();
									$qualifier          = "";
									foreach ($itemExtraFieldData as $CRDataNodeId_lookup)
									{
										$qualifier .= "dataNodeID={$CRDataNodeId_lookup}";
										if ($CRDataNodeId_lookup !== end($itemExtraFieldData)) $qualifier .= " or ";
									}

									$params = array(
										'qualifier' => $qualifier,
										'show'      => array('value'),
										'language'  => $cx_forced_lang_name ?: $lang_obj["name"],
										'_cxwp'     => Carerix::VERSION
									);

									if (count($itemExtraFieldData))
									{
										$arrData = CRDataNode::findAll("list", $params)->getArrayCopy();
									}
									else
									{
										$arrData = array(); // everything is unchecked
									}

									$arrOptionValues = array();
									/* @var $CRDataNode CRDataNode */
									foreach ($arrData as $CRDataNode)
									{
										$arrOptionValues[] = $CRDataNode->getValue();
									}

									$fieldvalue = $arrOptionValues;
									break;

								case "hiddenType":
									$fieldvalue[] = WP_Constants::ADDITIONAL_FIELD_DELETED_KEY;
									break;

								default: // unsupported
									$fieldvalue[] = WP_Constants::ADDITIONAL_FIELD_DELETED_KEY;
							}

							list ($wp_fieldname_gen, $wp_fieldname) = $this->__createWpFieldNames($CRDataNodeId, (string) $exportCode);

							/**
							 * Rule: publication supercedes vacancy
							 * So if the current field holds data we can unset a possible previous entry (which might be the vacancy entity)
							 */
							if (count($fieldvalue)) unset($additionalInfo->{$wp_fieldname});
							foreach ($fieldvalue as $value_string) {
								$additionalInfo->addChild($wp_fieldname, $value_string);
							}

							/**
							 * Rule: if an additional field contains 2 or more entries and one of them is the
							 * "___::DELETED!!___" identifier, remove that identifier.
							 * Reason: if there are 2 identical exportcodes defined in an additional field for a publication and vacancy,
							 * the disabled entity (by the "isHidden" property) may not disable the other one.
							 */
							if (count($additionalInfo->{$wp_fieldname}) >= 2) {
								$pos = array_search(WP_Constants::ADDITIONAL_FIELD_DELETED_KEY, (array) $additionalInfo->{$wp_fieldname});
								if ($pos !== false) unset($additionalInfo->{$wp_fieldname}[$pos]);
							}

							/**
							 * If a fieldname based on the `exportcode` exists, try to remove the 'generated' one based on the CRNode ID
							 */
							if ($wp_fieldname != $wp_fieldname_gen) {
								$additionalInfo->addChild($wp_fieldname_gen, WP_Constants::ADDITIONAL_FIELD_DELETED_KEY);
							}
						} // foreach
					} // foreach

					// Finally, we need to delete additional fields which values are "cleared/voided" in CX. Problem is that empty fields aren't pushed
					foreach ($arrExtraFieldsDictionary as $CRDataNode)
					{
						$CRDataNodeId = $CRDataNode->getDataNodeID();
						$exportCode   = $CRDataNode->getExportCode();
						list ($wp_fieldname_gen, $wp_fieldname) = $this->__createWpFieldNames($CRDataNodeId, (string) $exportCode);
						if (!isset($additionalInfo->{$wp_fieldname_gen}))
						{
							$additionalInfo->addChild($wp_fieldname_gen, WP_Constants::ADDITIONAL_FIELD_DELETED_KEY);
						}
						if ($wp_fieldname != $wp_fieldname_gen && !isset($additionalInfo->{$wp_fieldname}))
						{
							$additionalInfo->addChild($wp_fieldname, WP_Constants::ADDITIONAL_FIELD_DELETED_KEY);
						}
					}
				}


				// Latest modification date determined by the publication AND related entities (like vacancy, groups, educations, functionlevels, etc)
				// (leave it here at the last position, else the timestamp might be incorrect)
				$modificationDate = $this->getCxTablesLastModificationDate();
				$thisItem->addChild("modificationDate", $modificationDate);

				$publication = null;
				unset($publication);
			}

			$xmlPublications = null;
			unset($xmlPublications);
			$arrPublications = null;
			unset($arrPublications);
			$data = null;
			unset($data);
			$manager = null;
			unset($manager);
			$client = null;
			unset($client);

			return $feedData;

		} catch (Carerix\Api\Rest\Exception $care) {
			die();
		}
	}



	public function process($data) 
	{
        Utils_Cache::cleanEnvironment();
		return $this->processData($data);
	}

    public function getCategory() 
    {
        // get the parent category for this jobs source based on medium
        $jobsUrl = $this->getUrl();
        $data = parse_url($jobsUrl);
        $query = array();
        $category = false;
        if(isset($data['query'])) 
        {
            parse_str($data['query'], $query);
        }
        
        if(isset($query['medium']) && $query['medium'] !== 'web' && stripos($query['medium'], 'web') !==0 ) 
        {
            if(stripos($query['medium'], '-') !== FALSE)
            {
                $v = explode('-', $query['medium']);
                $category = $v[0];
            }
            else
            {
                $category = $query['medium'];
            }
        }
        
        if(isset($query['medium']) && $query['medium'] == 'web')
        {
            $category = 'Publications';
        }
        
        return $category;
    }
    
	private function processData($data): int
	{
		$items = $data->channel->item;
		$itemsTable = $this->getArrayOfItems($items);
        $customFields = $this->getCustomFieldsData($items);

		$listOfFragments = array();
        $category = $this->getCategory();

		do_action( 'cxwordpress_sync_jobs_start', count($itemsTable) );

		/**
		 * for each item in the table,
		 * create a writer,
		 * write,
		 * collect a fragment
		 */
        $idx = 0;
		foreach($itemsTable as $item)
		{
			/** @noinspection PhpUnhandledExceptionInspection */
			$job = new Entity_Writer_Job($item, $customFields[$idx]);
            $job->setSourceData($this->sourceData);
			if($category) $job->setCategoryKey ($category);
			if($this->registerLanguage !== FALSE)
			{
				$job->setCategoryLanguage($this->registerLanguage);
				$job->setTypeSuffix($this->registerLanguage);
			}

			$id = $job->write();

			array_push($listOfFragments, $job->getFragment($id));
			$this->interfacer->entity_set_property($id,'source', $this->sourceData['id']);
            $this->interfacer->entity_set_property($id, 'source_cron', $this->sourceData['id']);

			/**
			 * register this post for the set language in the WPML table
			 */
			if($this->registerLanguage !== FALSE)
			{
				$trid = trim($item['vacancyID']);
				$this->interfacer->wpml_register_translation($id, $trid, 'post_post', $this->registerLanguage);
			}
            $idx++;
            $job = null;
		}

		do_action( 'cxwordpress_sync_jobs_finish', count($itemsTable) );

        return count($itemsTable);
	}
    
    public function setPostID($id) {
        $this->post_id = $id;
    }
    
    public function getPostID() {
        return $this->post_id;
    }
	
	/**
	 * Formats the retrieved xml into a list of tables
	 * 
	 * 
	 * @param SimpleXML $data
	 * @return array
	 */
	private function getArrayOfItems($data)
	{
		$items = array();
		
		$collectors = array(
                    'guid', 'pubDate', 'endDate', 'deadlineMaterial', 'reactionEndDate', 'title',
                    'introInformation',  // 1
                    'companyInformation',  // 2
                    'vacancyInformation', // 3
                    'requirementsInformation', // 4
                    'offerInformation', // 5
                    'functionContactInformation', // 6
                    'applicationContactInformation', // 7
                    'category',   'workLocation',
                    'rawTitle', 'titleWorkLocation',
                    'vacancyID', 'country1', 'country2', 'country3',
                    'region1', 'region2', 'region3', 'medium', 'educations', 'procedure', 'businessLine',
                    'branche0', 'branche1',
                    'function0', 'function1',
                    'toWorkLevelNode',
                    'ownerName',
					'contractType',
					'groups',
					'hoursPerWeek',
					'isSalaryPublic', 'minSalary', 'maxSalary', 'salaryPeriod', 'salaryPeriodClassification', 'salaryCurrency',
					'pubIdList',
		);
		
        
		foreach($data as $item)
		{
			$newItem = array();
			
			foreach($collectors as $collector)
			{
                if($collector == 'educations')
                {
                    $educations = array();
                    foreach($item->$collector->education as $education)
                    {
                        $education = (string)$education;
                        array_push($educations, $education);
                    }
                    $newItem[$collector] = $educations;
                    
                    $requiredEducations = array();
                    foreach($item->$collector->required as $requiredEducation)
                    {
                        $required = (string)$requiredEducation;
                        array_push($requiredEducations, $required);
                    }
                    $newItem['requiredEducations'] = $requiredEducations;
                }

                // vacancy groups
                if($collector == 'groups')
                {
                    $groups = array();
//	                $groupsId = array();
	                $groupsExportCode = array();
                    foreach($item->$collector->group as $group)
                    {
                        $groupname = (string)$group;
                        array_push($groups, $groupname);
//                      array_push($groupsId, $group['id']); // get the attribute
                        array_push($groupsExportCode, $group['exportCode']); // get the attribute
                    }
                    $newItem[$collector] = $groups;
//	                $newItem['id'] = $groupsId;
	                $newItem['groupsExportCode'] = $groupsExportCode;
                }

				if(isset($item->$collector)
                    && strlen(trim((string)$item->$collector)) > 0 
                    )
				{
                    $name_collector = preg_replace('/\s+/u'  , ' | ',trim(strip_tags((string)$item->$collector)));
                    $name_collector = htmlentities($name_collector, null, 'utf-8');
                    $name_collector = str_replace('&nbsp;', '', $name_collector);
                    $name_collector = str_replace('&#160;', '', $name_collector); 

                    if(strlen($name_collector) == 0) continue;
                    
					$newItem[$collector] = preg_replace('/\s+/u', ' ', trim((string)$item->$collector));
                    
                    if($collector == 'title') {
                        $newItem[$collector] = preg_replace('/\s+/u', ' ', trim((string)$item->rawTitle));
                        $typeOfTitle = @$this->sourceData['settings']['title_compose'];
                        if($typeOfTitle == 'title') {
                            $real_title = $newItem[$collector];
                        } else {
                            $location = preg_replace('/\s+/u', ' ', trim((string)$item->titleWorkLocation));
                            $real_title = implode(' - ', array($newItem[$collector], $location));
                        }
                        $newItem[$collector] = $real_title;
                    }
                    
				}
			}
			
			if(!empty($newItem)) array_push ($items, $newItem);
		}
		
        
		return $items;
	}

	/**
	 * Retrieves the CX fields from the xml feed to store as metadata. (Metadata aren't taxonomies)
	 * These fields are retrievable as: get_post_meta(get_the_ID(), 'postalCode', true);
	 * The original XML fieldnames have to be used.
	 *
	 * @param $data
	 *
	 * @return array
	 */
	private function getCustomFieldsData($data) {
        
        $customFieldCollectors = array(
	        'medium', 'guid', 'vacancyID', 'vacancyNo',
            'pubDate', 'startDate', 'endDate', 'deadlineMaterial', 'reactionEndDate',
	        'branche0', 'branche1', 'function0', 'function1',
	        'title', 'workLocation', 'contact', 'company', 'postalCode',
	        'introInformation', 'companyInformation', 'vacancyInformation',
            'requirementsInformation', 'offerInformation', 'functionContactInformation',
            'applicationContactInformation',
	        'contractType', 'hoursPerWeek',
	        'isSalaryPublic', 'minSalary', 'maxSalary', 'salaryPeriod', 'salaryPeriodClassification', 'salaryCurrency',
	        'pubIdList', 'procedure', 'businessLine',
        );
        $items = array(); // batch of vacancies
        
        foreach($data as $item) {
            $newItem = array();
            foreach($customFieldCollectors as $collector) {
                if(isset($item->$collector) /*&& strlen(trim((string)$item->$collector)) > 0*/ ) // also update voided CX fields, like 'endDate'
                {
                    $newItem[$collector] = preg_replace('/\s+/u', ' ', trim((string)$item->$collector));
                }
                else {
	                $newItem[$collector] = ""; // Force Custom Field (= meta data) to an empty string. This happens when new REST-API fields are used by the deprecated cxtools wordpress xml feed
                }
            }

			// Process the optional additional cx fields
	        // If this functionality is turned off in the plugin settings, it will hold an empty array
	        $cx_additionalFields = (array) $item->additionalInfo;

	        foreach ($cx_additionalFields as $cx_field_key=>$cx_field_value) {
		        $newItem[$cx_field_key] = $cx_field_value;
			}
            if(!empty($newItem)) array_push ($items, $newItem);
        }

        return $items;
    }


	/**
	 * Determines whether a vacancy is set to anonymous to hide sensitive information
	 *
	 * @param $publication
	 *
	 * @return bool
	 */
	private function publicationIsAnonymous( $publication )
	{
		$isAnonymous = $publication->getToVacancy()->isAnonymous;
		if ($isAnonymous != 1) {
			return false;
		} else {
			return true;
		}
	}


	/**
	 * Travels from the current publication to its vacancy object, then descends to all connected publications.
	 * Finally returns a comma delimited list of all the pubIDs.
	 *
	 * @param $publication
	 *
	 * @return string
	 */
	private function getOtherPubIdsBelongingToThisPublication( $publication )
	{
		$publicationMappings = $publication->getToVacancy()->getPublications();
		$arrPubIds = array();
		foreach ($publicationMappings as $mappedPublication)
		{
			$arrPubIds[] = $mappedPublication->publicationID;
		}

		return implode(",", $arrPubIds );
	}

	/**
	 * Convert "2019-02-14 12:00:00" to something like "Thu, 14 Feb 2019 12:00:00 +0100"
	 *
	 * @param $inputDateTime
	 *
	 * @return string
	 */
	private function formatPublicationDate( $inputDateTime )
	{
		$format = 'Y-m-d H:i:s';
		$date = DateTime::createFromFormat($format, $inputDateTime);
		$timezone_as_emailformat = $this->formatTimeZoneAsEmail(); // something like +0100 or +0200
		return $date->format('D, d M Y H:i:s') . " {$timezone_as_emailformat}";
	}

	/**
	 * Format job title to one including working location if provided (or in some cases when allowed)
	 *
	 * @param $publication
	 *
	 * @return string
	 */
	private function formatPublicationTitle( $publication )
	{
		$titleInformation   = trim($publication->getTitleInformation());
		$workLocation       = trim($publication->getWorkLocation());
		$workCity           = trim($publication->getToVacancy()->workCity);
		$invoicePostalCode  = trim($publication->getToVacancy()->invoicePostalCode);
		$visitPostalCode    = trim($publication->getToVacancy()->getToCompany()->visitPostalCode);
		$isAnonymous        = $this->publicationIsAnonymous( $publication );

		$title = $titleInformation;
		if (strlen($workLocation)) {
			$title = "{$title} - {$workLocation}";
		} elseif (strlen($workCity)) {
			$title = "{$title} - {$workCity}";
		} elseif (strlen($invoicePostalCode)) {
			$title = "{$title} - {$invoicePostalCode}";
		} elseif ( $isAnonymous == false && strlen($visitPostalCode)) {
			$title = "{$title} - {$visitPostalCode}";
		}

		return htmlspecialchars($title);
	}

	/**
	 * Return the timezone offset (e.g. +0100) for Europe/Amsterdam. CX uses this timezone internally
	 *
	 * @return string
	 */
	private function formatTimeZoneAsEmail()
	{
		$tz                         = timezone_open("Europe/Amsterdam");
		$dateTimeUTC0               = date_create("now",timezone_open("UTC"));
		$diff_hours                 = timezone_offset_get($tz,$dateTimeUTC0) / (60*60);
		$timezone_as_emailformat    = sprintf('%+03d00', $diff_hours); // something like: +0100 or +0200
		return $timezone_as_emailformat;
	}


	/**
	 * Get work location if provided (or in some cases when allowed)
	 *
	 * @param $publication
	 *
	 * @return string
	 */
	private function getTitleWorkLocation( $publication )
	{
		$workLocation       = trim($publication->getWorkLocation());
		$workCity           = trim($publication->getToVacancy()->workCity);
		$invoicePostalCode  = trim($publication->getToVacancy()->invoicePostalCode);
		$visitPostalCode    = trim($publication->getToVacancy()->getToCompany()->visitPostalCode);
		$isAnonymous        = $this->publicationIsAnonymous( $publication );

		$titleWorkLocation = "";
		if (strlen($workLocation)) {
			$titleWorkLocation = $workLocation;
		} elseif (strlen($workCity)) {
			$titleWorkLocation = $workCity;
		} elseif (strlen($invoicePostalCode)) {
			$titleWorkLocation = $invoicePostalCode;
		} elseif ( $isAnonymous == false && strlen($visitPostalCode)) {
			$titleWorkLocation = $visitPostalCode;
		}

		return htmlspecialchars($titleWorkLocation);
	}

	/**
	 * Fetch the groups
	 *
	 * @param SimpleXMLElement $publicationXML
	 *
	 * @return array
	 */
	private function getPublicationGroups( $publicationXML )
	{
		$groupsXML = $publicationXML->toVacancy->CRVacancy->groups;
		$group_count = $groupsXML->attributes()->count;

		$groups = array();
		for ($i=0; $i<$group_count; $i++)
		{
			$groupXML   = $groupsXML->CRDataNode[$i];
			$value      = isset($groupXML->value) ? (string) $groupXML->value : "";
			$exportCode = isset($groupXML->exportCode) ? strip_tags($groupXML->exportCode) : "";
			$dataNodeID = isset($groupXML->dataNodeID) ? (string) $groupXML->dataNodeID : "";
			$skip       = 0;
			if ( isset($groupXML->isHidden) && $groupXML->isHidden == 1 ) $skip = 1;
			if ( isset($groupXML->deleted) && $groupXML->deleted == 1 ) $skip = 1;

			$modificationDate = isset($groupXML->modificationDate) ? (string) $groupXML->modificationDate : $this->interfacer::INIT_SYNC_STARTDATE;
			$this->__addModificationDate( $modificationDate );
			if ( $skip ) continue;
			$groups[] = array(
				"value" => $value,
				"exportCode" => $exportCode,
				"id" => $dataNodeID,
				"modificationDate" => $modificationDate,
			);
		}

		return $groups;
	}


	/**
	 * Fetch educations
	 *
	 * @param SimpleXMLElement $publicationXML
	 *
	 * @return array
	 */
	private function getEducations( $publicationXML )
	{
		$educations_XML = $publicationXML->toVacancy->CRVacancy->educations;
		$education_count = $educations_XML->attributes()->count;

		$educations = array();

		for ($i=0; $i<$education_count; $i++)
		{
			$CRVacancyEducation   = $educations_XML->CRVacancyEducation[$i];

			$toLevel1Education1 = isset($CRVacancyEducation->toLevel1Education1->CRDataNode->value) ? (string) $CRVacancyEducation->toLevel1Education1->CRDataNode->value : "";
			$modificationDate = isset($CRVacancyEducation->toLevel1Education1->CRDataNode->modificationDate) ? (string) $CRVacancyEducation->toLevel1Education1->CRDataNode->modificationDate : $this->interfacer::INIT_SYNC_STARTDATE;
			$this->__addModificationDate( $modificationDate );
			$toLevel1Education2 = isset($CRVacancyEducation->toLevel1Education2->CRDataNode->value) ? (string) $CRVacancyEducation->toLevel1Education2->CRDataNode->value : "";
			$modificationDate = isset($CRVacancyEducation->toLevel1Education2->CRDataNode->modificationDate) ? (string) $CRVacancyEducation->toLevel1Education2->CRDataNode->modificationDate : $this->interfacer::INIT_SYNC_STARTDATE;
			$this->__addModificationDate( $modificationDate );
			$toLevel1Education3 = isset($CRVacancyEducation->toLevel1Education3->CRDataNode->value) ? (string) $CRVacancyEducation->toLevel1Education3->CRDataNode->value : "";
			$modificationDate = isset($CRVacancyEducation->toLevel1Education3->CRDataNode->modificationDate) ? (string) $CRVacancyEducation->toLevel1Education3->CRDataNode->modificationDate : $this->interfacer::INIT_SYNC_STARTDATE;
			$this->__addModificationDate( $modificationDate );

			$educations[] = array(
				"education" => $toLevel1Education1,
				"required" => "{$toLevel1Education1} :|: {$toLevel1Education2} :|: {$toLevel1Education3}",
			);
		}

		return $educations;
	}

	/**
	 * Returns when last modification "in general" took place
	 *
	 * @return string
	 */
	private function getCxTablesLastModificationDate()
	{
		return $this->modified_datetime[0] ?? "2222-12-31 23:59:59";
	}

	/**
	 * Helper method to build an array of modification dates
	 * and returns the most recent date value.
	 *
	 * @param $datetime string Formatted as: 2050-07-15 15:00:00
	 *
	 * @return string Most recent date value
	 */
	private function __addModificationDate( $datetime )
	{
		if (isset ($datetime) ) {
			$this->modified_datetime[] = $datetime;
			rsort($this->modified_datetime);
		}
		return $this->modified_datetime[0];
	}


	private function __createWpFieldNames( int $CRDataNodeId, string $exportCode = "" ): array
	{
		$preferred_suffix   = trim(strtolower($exportCode));
		$preferred_suffix   = preg_replace("/[^a-z0-9]+/", "_", $preferred_suffix); // only allow a-z and digits. The rest is replaced by underscores.
		$preferred_suffix   = strlen($preferred_suffix) ? $preferred_suffix : $CRDataNodeId; // use 'exportcode' suffix, else fall back to node ID
		$wp_fieldname       = WP_Constants::ADDITIONAL_FIELD_PREFIX . $preferred_suffix;
		$wp_fieldname_gen   = WP_Constants::ADDITIONAL_FIELD_PREFIX . $CRDataNodeId; // generated fieldname
		return array($wp_fieldname_gen, $wp_fieldname);
	}

}
