<?php

/**
 * Carerix Wordpress Plugin
 *
 * @category settings
 * @author Bogdan Ivanov
 * @copyright Copyright (c) 2013 Carerix.com (http://www.carerix.com)
 * @link http://www.carerix.com
 * @since Sep 4, 2013 11:20:34 AM
 * @version $Rev$
 */


/**
 * Labels class
 * Used to ....
 */
class Utils_Labels {
	
	private static $alternates = array(
		'en-US' => array('en-GB'),
		'en'	=> array('en-GB'),
		'de'	=> array('de-DE'),
		'es'	=> array('es-ES'),
		'fr'	=> array('fr-FR'),
		'it'	=> array('it-IT'),
		'nl'	=> array('nl-NL'),
		'pl'	=> array('pl-PL'),
		'hu'	=> array('hu-HU'),
		'pt'	=> array('pt-PT'),
	);
	
	private function __construct() {
		;
	}
	
	static function get($language = FALSE)
	{
		if($language === FALSE)
		{
			$interfacer = WP_Interfacer::getInstance();
			$current_language = $interfacer->language_current();	
		}
		else
		{
			$current_language = trim($language);
		}
		
		$filenames = self::get_resource_file_name($current_language);
		if( count($filenames) == 0 )
		{
			$current_language = WP_Constants::DEFAULT_LANGUAGE_CODE;
			$filenames = self::get_resource_file_name($current_language);
		}
		
		return self::process_resources($filenames);
	}

	/**
	 * @param $language
	 *
	 * @return array    Array holds 2 filepath items if an override resource file is found in the (child)theme directory
	 */
	private static function get_resource_file_name($language)
	{
		$directory = CARERIX_INSTALL_DIR . DIRECTORY_SEPARATOR . 'assets' . DIRECTORY_SEPARATOR . 'resources';
		$childtheme_directory = get_stylesheet_directory() . DIRECTORY_SEPARATOR . plugin_basename( CARERIX_INSTALL_DIR ) . DIRECTORY_SEPARATOR  . 'assets' . DIRECTORY_SEPARATOR . 'resources';

		$resource_files = array();

		$filename = $language . '.resources.php';
		
		if(file_exists($directory . DIRECTORY_SEPARATOR . $filename)) $resource_files[] = $directory . DIRECTORY_SEPARATOR . $filename;
		if(file_exists($childtheme_directory . DIRECTORY_SEPARATOR . $filename)) $resource_files[] = $childtheme_directory . DIRECTORY_SEPARATOR . $filename;

		if ( count($resource_files) == 0 ) {
			$languageIndex = self::get_major_language($language);
			if(!$languageIndex) $languageIndex = 'en';
			if(!isset(self::$alternates[$languageIndex])) $languageIndex = 'en';

			foreach(self::$alternates[$languageIndex] as $alternate_language)
			{
				$filename = $alternate_language . '.resources.php';

				if(file_exists($directory . DIRECTORY_SEPARATOR . $filename)) $resource_files[] = $directory . DIRECTORY_SEPARATOR . $filename;
				if(file_exists($childtheme_directory . DIRECTORY_SEPARATOR . $filename)) $resource_files[] = $childtheme_directory . DIRECTORY_SEPARATOR . $filename;

				if ( count($resource_files) > 0 ) break;
			}
		}

		return $resource_files;
	}

	/**
	 * Processes the resources files for the particular language. If the input array contains 2 items the
	 * last file will override the earlier file, but only for the defined "resource keys".
	 * This makes it easier to maintain language overrides after updating the plugin.
	 *
	 * @param array $filename_paths
	 *
	 * @return array
	 */
	private static function process_resources($filename_paths)
	{
		$labels = array();

		foreach ($filename_paths as $filename)
		{
			$data = parse_ini_file($filename, true);
			foreach($data as $section => $values)
			{
				foreach($values as $key => $value)
				{
					$labels[$section . '.' . $key] = $value;
				}
			}
		}

		return $labels;
	}
	
	private static function get_major_language($language)
	{
		$separators = array('-', '_');
		
		foreach($separators as $separator)
		{
			if(stripos($language, $separator) !== false)
			{
				$_t = explode($separator,$language);
				return $_t[0];
			}			
		}
		return $language;
	}
}